@@ -1,9 +1,11 @@
 from typing import Tuple
 
 from django.contrib.auth.models import Group, User
+from django.contrib.humanize.templatetags.humanize import intcomma
 from django.core.exceptions import ValidationError
 from django.core.validators import MaxValueValidator, MinValueValidator
 from django.db import Error, models
+from django.utils.translation import gettext as _
 from esi.errors import TokenExpiredError, TokenInvalidError
 from esi.models import Token
 from eveuniverse.models import EveSolarSystem, EveType
@@ -14,7 +16,11 @@ from allianceauth.authentication.models import CharacterOwnership, State
 from allianceauth.eveonline.models import EveCorporationInfo
 from allianceauth.services.hooks import get_extension_logger
 
-from buybackprogram.app_settings import BUYBACKPROGRAM_TRACKING_PREFILL
+from buybackprogram.notification import (
+    send_message_to_discord_channel,
+    send_user_notification,
+)
+from buybackprogram.utils import get_site_url
 
 from .decorators import fetch_token_for_owner
 from .providers import esi
@@ -22,6 +28,15 @@ from .providers import esi
 logger = get_extension_logger(__name__)
 
 
+def get_sentinel_user():
+    """
+    get user or create one
+    :return:
+    """
+
+    return User.objects.get_or_create(username="deleted")[0]
+
+
 class General(models.Model):
     """Meta model for app permissions"""
 
@@ -53,9 +68,8 @@ class Owner(models.Model):
         (ERROR_ESI_UNAVAILABLE, "ESI API is currently unavailable"),
     ]
 
-    corporation = models.OneToOneField(
+    corporation = models.ForeignKey(
         EveCorporationInfo,
-        primary_key=True,
         on_delete=models.deletion.CASCADE,
         related_name="+",
     )
@@ -84,152 +98,202 @@ class Owner(models.Model):
     )
     def update_contracts_esi(self, token):
 
-        logger.debug("Fetching contracts for %s" % self.character)
+        logger.debug("Fetching contracts for %s" % self.user)
 
+        # Get all contracts for owner
         contracts = self._fetch_contracts()
 
-        for contract in contracts:
-
-            # Only get contracts with the correct prefill ticker
-            if BUYBACKPROGRAM_TRACKING_PREFILL in contract["title"]:
-                obj, created = Contract.objects.update_or_create(
-                    contract_id=contract["contract_id"],
-                    defaults={
-                        "contract_id": contract["contract_id"],
-                        "assignee_id": contract["assignee_id"],
-                        "availability": contract["availability"],
-                        "date_completed": contract["date_completed"],
-                        "date_expired": contract["date_expired"],
-                        "date_issued": contract["date_issued"],
-                        "for_corporation": contract["for_corporation"],
-                        "issuer_corporation_id": contract["issuer_corporation_id"],
-                        "issuer_id": contract["issuer_id"],
-                        "price": contract["price"],
-                        "status": contract["status"],
-                        "title": contract["title"],
-                        "volume": contract["volume"],
-                    },
-                )
-
-                # If we created an new contract
-                if created:
-                    logger.debug(
-                        "New contract %s created. Starting item fetch"
-                        % contract["contract_id"]
-                    )
-
-                    character_id = self.character.character.character_id
-
-                    contract_items = esi.client.Contracts.get_characters_character_id_contracts_contract_id_items(
-                        character_id=character_id,
-                        contract_id=contract["contract_id"],
-                        token=token.valid_access_token(),
-                    ).results()
-
-                    objs = []
-
-                    for item in contract_items:
-
-                        cont = Contract.objects.get(contract_id=contract["contract_id"])
-                        itm = EveType.objects.get(pk=item["type_id"])
-
-                        contract_item = ContractItem(
-                            contract=cont,
-                            eve_type=itm,
-                            quantity=item["quantity"],
-                        )
-
-                        objs.append(contract_item)
-
-                    try:
-                        ContractItem.objects.bulk_create(objs)
-                        logger.debug(
-                            "Succesfully added items for contract %s into database"
-                            % contract["contract_id"]
-                        )
-                    except Error as e:
-                        logger.error(
-                            "Error adding items for contract %s: %s"
-                            % (contract["contract_id"], e)
-                        )
-
-                    logger.debug(
-                        "Updated items for contract %s" % contract["contract_id"]
-                    )
-                else:
-                    logger.debug("Contract %s updated." % contract["contract_id"])
-
-        logger.debug("Fetching corporation contracts for %s" % self.character)
+        logger.debug("Got %s character contracts" % len(contracts))
 
+        # Get all contracts for owner corporation
         corporation_contracts = self._fetch_corporation_contracts()
 
-        for contract in corporation_contracts:
+        logger.debug("Got %s corporation contracts" % len(corporation_contracts))
 
-            # Only get contracts with the correct prefill ticker
-            if BUYBACKPROGRAM_TRACKING_PREFILL in contract["title"]:
-                obj, created = Contract.objects.update_or_create(
-                    contract_id=contract["contract_id"],
-                    defaults={
-                        "contract_id": contract["contract_id"],
-                        "assignee_id": contract["assignee_id"],
-                        "availability": contract["availability"],
-                        "date_completed": contract["date_completed"],
-                        "date_expired": contract["date_expired"],
-                        "date_issued": contract["date_issued"],
-                        "for_corporation": contract["for_corporation"],
-                        "issuer_corporation_id": contract["issuer_corporation_id"],
-                        "issuer_id": contract["issuer_id"],
-                        "price": contract["price"],
-                        "status": contract["status"],
-                        "title": contract["title"],
-                        "volume": contract["volume"],
-                    },
-                )
+        # Merge all contracts into a single list
+        all_contracts = contracts + corporation_contracts
+
+        logger.debug("Total contracts received: %s" % len(all_contracts))
+
+        # Get all tracking objects from the database
+        tracking_numbers = Tracking.objects.all()
+
+        # Start looping for all stored tracking objects
+        for tracking in tracking_numbers:
+            # Start checking if we find any matches from our ESI contracts
+            for contract in all_contracts:
+
+                # Only get contracts with the correct prefill ticker
+                if tracking.tracking_number in contract["title"]:
 
-                if created:
                     logger.debug(
-                        "New corporation contract %s created. Starting item fetch"
+                        "Found a matching contract from tracking for %s "
                         % contract["contract_id"]
                     )
 
-                    corporation_id = self.character.character.corporation_id
-
-                    contract_items = esi.client.Contracts.get_corporations_corporation_id_contracts_contract_id_items(
-                        corporation_id=corporation_id,
-                        contract_id=contract["contract_id"],
-                        token=token.valid_access_token(),
-                    ).results()
-
-                    objs = []
-
-                    for item in contract_items:
-
-                        cont = Contract.objects.get(contract_id=contract["contract_id"])
-                        itm = EveType.objects.get(pk=item["type_id"])
-
-                        contract_item = ContractItem(
-                            contract=cont,
-                            eve_type=itm,
-                            quantity=item["quantity"],
-                        )
-
-                        objs.append(contract_item)
-
+                    # Check if we already have the contract stored
                     try:
-                        ContractItem.objects.bulk_create(objs)
-                        logger.debug(
-                            "Succesfully added items for contract %s into database"
-                            % contract["contract_id"]
+                        old_contract = Contract.objects.get(
+                            contract_id=contract["contract_id"]
                         )
-                    except Error as e:
-                        logger.error(
-                            "Error adding items for contract %s: %s"
-                            % (contract["contract_id"], e)
+
+                    except Contract.DoesNotExist:
+                        logger.debug("No matching contracts found")
+                        old_contract = Contract.objects.none()
+                        old_contract.status = False
+
+                    # If we have found a contract from database that is not yet finished
+                    if old_contract.status not in ["finished", "rejected"]:
+
+                        # Create or update the found contract
+                        obj, created = Contract.objects.update_or_create(
+                            contract_id=contract["contract_id"],
+                            defaults={
+                                "contract_id": contract["contract_id"],
+                                "assignee_id": contract["assignee_id"],
+                                "availability": contract["availability"],
+                                "date_completed": contract["date_completed"],
+                                "date_expired": contract["date_expired"],
+                                "date_issued": contract["date_issued"],
+                                "for_corporation": contract["for_corporation"],
+                                "issuer_corporation_id": contract[
+                                    "issuer_corporation_id"
+                                ],
+                                "issuer_id": contract["issuer_id"],
+                                "start_location_id": contract["start_location_id"],
+                                "price": contract["price"],
+                                "status": contract["status"],
+                                "title": contract["title"],
+                                "volume": contract["volume"],
+                            },
                         )
-                else:
-                    logger.debug(
-                        "Corporation contract %s updated." % contract["contract_id"]
-                    )
+
+                        # If we have created a new contract
+                        if created:
+                            logger.debug(
+                                "New contract %s created. Starting item fetch"
+                                % contract["contract_id"]
+                            )
+
+                            character_id = self.character.character.character_id
+
+                            # Get all items in the contract
+                            contract_items = esi.client.Contracts.get_characters_character_id_contracts_contract_id_items(
+                                character_id=character_id,
+                                contract_id=contract["contract_id"],
+                                token=token.valid_access_token(),
+                            ).results()
+
+                            objs = []
+
+                            # Prepare objects for bulk create
+                            for item in contract_items:
+
+                                cont = Contract.objects.get(
+                                    contract_id=contract["contract_id"]
+                                )
+                                itm = EveType.objects.get(pk=item["type_id"])
+
+                                contract_item = ContractItem(
+                                    contract=cont,
+                                    eve_type=itm,
+                                    quantity=item["quantity"],
+                                )
+
+                                objs.append(contract_item)
+
+                            try:
+                                ContractItem.objects.bulk_create(objs)
+                                logger.debug(
+                                    "Succesfully added items for contract %s into database"
+                                    % contract["contract_id"]
+                                )
+                            except Error as e:
+                                logger.error(
+                                    "Error adding items for contract %s: %s"
+                                    % (contract["contract_id"], e)
+                                )
+
+                            """
+                            Notifications for users who have the notifications enabled
+                            """
+
+                            disablation_hint = "*You can disable these notifications from your program settings*"
+
+                            user_message = {
+                                "title": "New contract assigned",
+                                "description": "A new buyback contract {0} with a value of {1} ISK has been assigned to you.\n\n{2}".format(
+                                    tracking.tracking_number,
+                                    intcomma(int(contract["price"])),
+                                    disablation_hint,
+                                ),
+                                "footer": "AA Buyback Program",
+                            }
+
+                            # If tracking is active and we should send a message for our users
+                            if tracking.program.discord_dm_notification:
+
+                                send_user_notification(
+                                    user=self.user,
+                                    level="success",
+                                    message=user_message,
+                                )
+
+                            # Notifications for the discord channel
+                            if tracking.program.discord_channel_notification:
+                                send_message_to_discord_channel(
+                                    channel_id=tracking.program.discord_channel_notification,
+                                    message=user_message,
+                                )
+
+                        # If contract was updated instead of created
+                        else:
+                            logger.debug("Contract %s updated." % obj.contract_id)
+
+                        # Check if the contract status has changed from ongoing to finished.
+                        if (
+                            old_contract.status == "outstanding"
+                            and obj.status == "finished"
+                        ):
+                            # Get notification settings for the contract issuer
+                            user_settings = UserSettings.objects.get(
+                                user=tracking.issuer_user
+                            )
+
+                            # If user has not disabled notifications
+                            if user_settings.disable_notifications is False:
+
+                                disablation_hint = "*You can disable these notifications from [My Settings]({0}/buybackprogram/user_settings_edit)*".format(
+                                    get_site_url()
+                                )
+
+                                # Check if the contract was accepted or rejected
+                                if contract["status"] == "finished":
+                                    status = "accepted"
+                                elif contract["status"] == "rejected":
+                                    status = "rejected"
+                                else:
+                                    status = contract["status"]
+
+                                user_message = {
+                                    "title": "Buyback contract {0}".format(status),
+                                    "description": "Your outstanding buyback contract {0} with a value of {1} ISK has been {2} by {3}.\n\n{4}".format(
+                                        tracking.tracking_number,
+                                        intcomma(int(contract["price"])),
+                                        status,
+                                        self.character.character,
+                                        disablation_hint,
+                                    ),
+                                    "footer": "AA Buyback Program",
+                                }
+
+                                send_user_notification(
+                                    user=tracking.issuer_user,
+                                    level="success",
+                                    message=user_message,
+                                )
+
+                    break  # If we have found a match from our ESI contracts wi will stop looping on the contracts
 
     @fetch_token_for_owner(["esi-contracts.read_character_contracts.v1"])
     def _fetch_contracts(self, token) -> list:
@@ -326,6 +390,7 @@ class Location(models.Model):
 
     eve_solar_system = models.ForeignKey(
         EveSolarSystem,
+        verbose_name="Solar system",
         help_text="System where the buyback structure is located",
         blank=True,
         default=None,
@@ -334,8 +399,31 @@ class Location(models.Model):
         related_name="+",
     )
 
+    owner = models.ForeignKey(
+        Owner,
+        verbose_name="Manager",
+        help_text="Player managing this location",
+        null=True,
+        on_delete=models.deletion.CASCADE,
+    )
+
+    structure_id = models.BigIntegerField(
+        verbose_name="Ingame unique ID for structure",
+        default=None,
+        blank=True,
+        null=True,
+        help_text="The ID for the structure you wish to accept the contracts at. If left empty the program statistics page will not track if the contract is actually made at the correct structure or not. To get the ID for the structure see readme for getting structure IDs",
+    )
+
     def __str__(self):
-        return self.eve_solar_system.name + " | " + self.name
+
+        return (
+            self.eve_solar_system.name
+            + " | "
+            + self.name
+            + ", ID: "
+            + str(self.structure_id)
+        )
 
 
 class Program(models.Model):
@@ -343,51 +431,57 @@ class Program(models.Model):
 
     owner = models.ForeignKey(
         Owner,
-        help_text="Player that the contracts will be created for",
+        verbose_name="Manager",
+        help_text="Character that is used to manage this program.",
         on_delete=models.deletion.CASCADE,
         related_name="+",
     )
 
     is_corporation = models.BooleanField(
         default=False,
-        help_text="Tick if contracts should be made to corporation instead of player.",
+        help_text="If we should use the corporation of the manager as the contract receiver instead of the character.",
     )
 
     location = models.ForeignKey(
         Location,
-        help_text="Solarystem and station name for contracts.",
+        help_text="The location where contracts should be created at.",
         on_delete=models.deletion.CASCADE,
         related_name="+",
     )
 
     tax = models.IntegerField(
+        verbose_name="Default tax",
         default=0,
         blank=False,
         null=False,
-        help_text="Default tax is applied on all items unless an item spesific tax is assigned",
+        help_text="A default tax rate in this program that is applied on all items.",
         validators=[MaxValueValidator(100), MinValueValidator(0)],
     )
 
     hauling_fuel_cost = models.IntegerField(
+        verbose_name="Hauling fuel cost per m³",
         default=0,
         help_text="ISK per m³ that will be removed from the buy price ie. to cover jump freighet fuel costs. <b>Should not be used with price dencity modifier</b>",
     )
 
     price_dencity_modifier = models.BooleanField(
+        verbose_name="Price density modifier",
         default=False,
         help_text="Should we modify buy prices for items with high volume and low value ie. T1 industrial hulls. <b>Should not be used with hauling fuel cost</b>",
     )
 
     price_dencity_treshold = models.IntegerField(
+        verbose_name="Price density threshold",
         default=0,
         null=True,
         help_text="At what ISK/m3 do we start to apply the low isk dencity tax. Tritanium is 500 ISK/m³ @ 5 ISK per unit price. PLEX is 14,5Trillion ISK/m³ @2.9M per unit price.",
     )
 
     price_dencity_tax = models.IntegerField(
+        verbose_name="Price density tax",
         default=0,
         null=True,
-        help_text="How much tax do we apply on the low isk dencity items.",
+        help_text="How much tax do we apply on the low isk density items.",
         validators=[MaxValueValidator(100), MinValueValidator(0)],
     )
 
@@ -397,32 +491,55 @@ class Program(models.Model):
     )
 
     use_refined_value = models.BooleanField(
+        verbose_name="Ore: Use refined value",
         default=False,
         help_text="Take refined value into account when calculating prices for ore, ice and moon goo",
     )
 
     use_compressed_value = models.BooleanField(
+        verbose_name="Ore: Use compressed value",
         default=False,
         help_text="Take compressed value into account when calculating prices for ore, ice and moon goo",
     )
 
     use_raw_ore_value = models.BooleanField(
+        verbose_name="Ore: Use raw value",
         default=True,
         help_text="Take raw ore value into account when calculating prices for ore, ice and moon goo",
     )
 
     allow_unpacked_items = models.BooleanField(
+        verbose_name="Allow unpacked items",
         default=False,
         help_text="Do you want to allow unpacked items in this program such as assembled ship hulls?",
     )
 
     refining_rate = models.IntegerField(
+        verbose_name="Refining rate",
         default=0,
         null=True,
-        help_text="Refining rate to be used if refined value is active",
+        help_text="Refining rate to be used if ore refined value is active",
         validators=[MaxValueValidator(100), MinValueValidator(0)],
     )
 
+    blue_loot_npc_price = models.BooleanField(
+        verbose_name="NPC price for: Sleeper Components",
+        default=False,
+        help_text="Use NPC price as value for blue loot",
+    )
+
+    red_loot_npc_price = models.BooleanField(
+        verbose_name="NPC price for: Triglavian Survey Database",
+        default=False,
+        help_text="Use NPC price as value for red loot",
+    )
+
+    ope_npc_price = models.BooleanField(
+        verbose_name="NPC price for: Overseer's Personal Effects",
+        default=False,
+        help_text="Use NPC price as value for OPEs",
+    )
+
     restricted_to_group = models.ManyToManyField(
         Group,
         blank=True,
@@ -436,6 +553,19 @@ class Program(models.Model):
         help_text="The state(s) that will be able to see this buyback program. If none is selected program is open for all.",
     )
 
+    discord_dm_notification = models.BooleanField(
+        verbose_name="Discord direct messages for new contracts",
+        default=False,
+        help_text="Check if you want to receive a direct message notification each time a new contract is created. <b>Requires aa-discordbot app or discordproxy app to work</b>",
+    )
+
+    discord_channel_notification = models.BigIntegerField(
+        verbose_name="Discord channel ID for notifications",
+        null=True,
+        blank=True,
+        help_text="Check if you want to feed new contracts to a discord channel. <b>Requires aa-discordbot app or discordproxy app to work</b>",
+    )
+
     class Meta:
         default_permissions = ()
 
@@ -478,13 +608,15 @@ class ProgramItem(models.Model):
         help_text="Select item for special tax",
     )
     item_tax = models.IntegerField(
+        verbose_name="Item tax adjustment",
         default=0,
         null=True,
-        help_text="How much tax do we add on top of the base tax for this item. Can also be negative.",
-        validators=[MaxValueValidator(100), MinValueValidator(0)],
+        help_text="How much do you want to adjust the default tax on this item. Can be a positive or a negative value.",
+        validators=[MaxValueValidator(100), MinValueValidator(-100)],
     )
 
     disallow_item = models.BooleanField(
+        verbose_name="Disallow item in program",
         default=False,
         help_text="You can disallow an item from a buyback location. It will return 0 price if disallowed.",
     )
@@ -560,6 +692,7 @@ class Contract(models.Model):
     for_corporation = models.BooleanField()
     issuer_corporation_id = models.IntegerField()
     issuer_id = models.IntegerField()
+    start_location_id = models.BigIntegerField(null=True)
     price = models.BigIntegerField()
     status = models.CharField(max_length=30)
     title = models.CharField(max_length=128)
@@ -581,3 +714,31 @@ class ContractItem(models.Model):
     )
 
     quantity = models.IntegerField()
+
+
+class UserSettings(models.Model):
+    """
+    User settings
+    """
+
+    user = models.ForeignKey(
+        User,
+        related_name="+",
+        null=True,
+        blank=True,
+        default=None,
+        on_delete=models.SET(get_sentinel_user),
+    )
+
+    disable_notifications = models.BooleanField(
+        default=False,
+    )
+
+    class Meta:
+        """
+        Meta definitions
+        """
+
+        default_permissions = ()
+        verbose_name = _("User Settings")
+        verbose_name_plural = _("User Settings")
